#define _GNU_SOURCE
#include "graphics.hh"


BImage::BImage(unsigned int w, unsigned int h, int d, unsigned long *i) {
  width = w;
  height = h;
  depth = d;

  bg_color.r = bg_color.g = bg_color.b = 0;
  unsigned int wh = width * height;
  if (! i) data =  new unsigned long[wh];
  else data = i;
  setBackgroundColor(bg_color.r, bg_color.g, bg_color.b);
}


BImage::BImage(unsigned int w, unsigned int h, int d, const BColor &c) {
  width = w;
  height = h;
  depth = d;

  unsigned int wh = width * height;
  data = new unsigned long [wh];
  setBackgroundColor(c);
}


BImage::BImage(unsigned int w, unsigned int h, int d, unsigned char r,
	       unsigned char g, unsigned char b)
{
  width = w;
  height = h;
  depth = d;

  unsigned int wh = width * height;
  data = new unsigned long[wh];
  setBackgroundColor(r, g, b);
}


BImage::BImage(XImage *img) {
  width = (unsigned int) img->width;
  height = (unsigned int) img->height;
  depth = img->depth;
}


BImage::~BImage(void) {
  if (data) delete [] data;
}


Bool BImage::getPixel(unsigned int x, unsigned int y, unsigned char *r,
		      unsigned char *g, unsigned char *b)
{
  if (x > width ||y > height) {
    *r = *g = *b = 0;
    return False;
  }

  unsigned int xy = x * y;
  unsigned long *p = data +xy;

  *r = (*p >> 16) & 0xff;
  *g = (*p >> 8) & 0xff;
  *b = (*p) & 0xff;

  return True;
}


Bool BImage::getPixel(unsigned int x, unsigned int y, unsigned long *pixel) {
  // NOTE:  pixel will be returned a 32bpp pixel value.

  if (x > width || y > height) {
    *pixel = 0;
    return False;
  }

  unsigned int xy = x * y;
  *pixel = *(data + xy);
  return True;
}


Bool BImage::putPixel(unsigned int x, unsigned int y, unsigned char r,
		      unsigned char g, unsigned char b)
{
  if (x > width || y > height) return False;
  
  unsigned int xy = x * y;
  *(data + xy) = (r << 16) + (g << 8) + b;
  return True;
}


Bool BImage::putPixel(unsigned int x, unsigned int y, unsigned long pixel) {
  // NOTE:  pixel must be a 32bpp pixel value.

  if (x > width || y > height) return False;

  unsigned int xy = x * y; 
  *(data + xy) = pixel;
  return True;
}



XImage *BImage::convertToXImage(Display *display, Visual *visual)
{
  char *d = new char[width * height * (depth / 8)];
  if (! d) return 0;

  XImage *image = XCreateImage(display, visual, depth, ZPixmap, 0, d, width,
			       height, 8, 0);
  
  if (image == NULL) return 0;

  unsigned char r, g, b;
  unsigned int wh = width * height;
  unsigned long *p = data;
    
  switch (depth) {
  case 32: break;

  case 24: {
    unsigned char *im = (unsigned char *) image->data;

    while (--wh) {
      *(im++) = (unsigned char) ((*p) & 0xff);
      *(im++) = (unsigned char) (((*p) & 0xff00) >> 8);
      *(im++) = (unsigned char) (((*p++) & 0xff0000) >> 16);
    }

    *(im++) = (unsigned char) ((*p) & 0xff);
    *(im++) = (unsigned char) (((*p) & 0xff00) >> 8);
    *(im) = (unsigned char) (((*p) & 0xff0000) >> 16);

    break; }

  case 16: {
    unsigned short *im = (unsigned short *) image->data;

    while (--wh) {
      r = (*p & 0xff0000) >> 16; g = (*p & 0xff00) >> 8; b = (*p++ & 0xff);
      *(im++) = (((r << 8) & image->red_mask)|
		 ((g << 3) & image->green_mask)|
		 ((b >>3) & image->blue_mask));
    }

    r = (*p & 0xff0000) >> 16; g = (*p & 0xff00) >> 8; b = (*p & 0xff);
    *(im) = (((r << 8) & image->red_mask)|
	     ((g << 3) & image->green_mask)|
	     ((b >> 3) & image->blue_mask));

    break; }

  case 15:
  case 8: {
    int rr, gg, bb;
    unsigned char *im = (unsigned char *) image->data;

    while (--wh) {
      rr = (((*(p) & 0xff0000) >> 16) * 6) / 0xff;
      gg = (((*(p) & 0xff00) >> 8) * 6) / 0xff;
      bb = ((*(p++) & 0xff) * 6) / 0xff;

      if (rr < 0) rr = 0;
      if (rr > 6) rr = 5;
      if (gg < 0) gg = 0;
      if (gg > 6) gg = 5;
      if (bb < 0) bb = 0;
      if (bb > 6) bb = 5;
      *(im++) = colors_8bpp[rr * 36 + gg * 6 + bb].pixel;
    }

    rr = (((*(p) & 0xff0000) >> 16) * 6) / 0xff;
    gg = (((*(p) & 0xff00) >> 8) * 6) / 0xff;
    bb = ((*(p) & 0xff) * 6) / 0xff;
    
    if (rr < 0) rr = 0;
    if (rr > 6) rr = 5;
    if (gg < 0) gg = 0;
    if (gg > 6) gg = 5;
    if (bb < 0) bb = 0;
    if (bb > 6) bb = 5;
    *(im) = colors_8bpp[rr * 36 + gg * 6 + bb].pixel;
    
    break; }
  }

  return image;
}


Pixmap BImage::convertToPixmap(Display *display, Visual *visual, Window window)
{
  Pixmap pixmap = XCreatePixmap(display, window, width, height,
			 (unsigned) depth);
  
  if (pixmap == None) return None;

  XImage *img = convertToXImage(display, visual);

  if (! img) {
    XFreePixmap(display, pixmap);
    return None;
  }

  XPutImage(display, pixmap, DefaultGC(display, DefaultScreen(display)),
	    img, 0, 0, 0, 0, width, height);

  XDestroyImage(img);
  return pixmap;
}


void BImage::setBackgroundColor(const BColor &c) {
  bg_color = c;

  unsigned int wh = width * height;
  unsigned long *p = data, pixel = ((c.r << 16)|(c.g << 8)|(c.b));
  while (--wh) *(p++) = pixel;
  *p = pixel;
}

  
void BImage::setBackgroundColor(unsigned char r, unsigned char g,
				unsigned char b)
{
  bg_color.r = r;
  bg_color.g = g;
  bg_color.b = b;

  unsigned int wh = width * height; 
  unsigned long *p = data, pixel = (r << 16) | (g << 8) | (b);

  while (--wh) *(p++) = pixel;
  *p = pixel;
}


void BImage::renderBevel(void) {
  if (height > 4) {
  unsigned long pix0, pix1;
  unsigned char r, g, b;

  r = bg_color.r * 5 / 4;
  if (r < bg_color.r) r = 0xff;
  g = bg_color.g * 5 / 4;
  if (g < bg_color.g) g = 0xff;
  b = bg_color.b * 5 / 4;
  if (b < bg_color.b) b = 0xff;
  
  pix0 = ((r << 16)|(g << 8)|b);
  
  r = bg_color.r * 3 / 4;
  if (r > bg_color.r) r = 0x0;
  g = bg_color.g * 3 / 4;
  if (g > bg_color.g) g = 0x0;
  b = bg_color.b * 3 / 4;
  if (b > bg_color.b) b = 0x0;

  pix1 = ((r << 16)|(g << 8)|b);

  // lets do the easy parts... top and bottom lines
  unsigned int w = width - 2, h = height - 1, wh = width * (height - 3);

  unsigned long *p = data + width + 1;

  while (--w) {
    *p = pix0;
    *((p++) + wh) = pix1;
  }

  p = data + width;
  while (--h) {
    *(++p) = pix0;
    p += (width - 3);
    *(p++) = pix1;
    p++;
  }
}
}


void BImage::renderButton(void) {
  unsigned long pix0, pix1;
  unsigned char r, g, b;

  r = bg_color.r * 5 / 4;
  if (r < bg_color.r) r = 0xff;
  g = bg_color.g * 5 / 4;
  if (g < bg_color.g) g = 0xff;
  b = bg_color.b * 5 / 4;
  if (b < bg_color.b) b = 0xff;
  
  pix0 = ((r << 16)|(g << 8)|b);
  
  r = bg_color.r * 3 / 4;
  if (r > bg_color.r) r = 0x0;
  g = bg_color.g * 3 / 4;
  if (g > bg_color.g) g = 0x0;
  b = bg_color.b * 3 / 4;
  if (b > bg_color.b) b = 0x0;

  pix1 = ((r << 16)|(g << 8)|b);

  // lets do the easy parts... top and bottom lines
  unsigned int w = width, h = height - 1, wh = w * h;

  unsigned long *p = data;

  while (--w) {
    *p = pix0;
    *((p++) + wh) = pix1;
  }
  *p = pix0;
  *(p + wh) = pix1;

  p = data + width;
  while (--h) {
    *p = pix0;
    p += (width - 1);
    *(p++) = pix1;
  }
  *p = pix0;
  p += (width - 1);
  *p = pix1;
}


void BImage::invertImage(void) {
  unsigned int wh = width * height;
  unsigned long *p, *pp, *new_data = new unsigned long[wh];
  pp = new_data;
  p = data + wh;

  while (--wh) *(pp++) = *(p--);
  *pp = *p;
  
  delete [] data;
  data = new_data;
}
